/* Copyright (C) 2008 Vin Helminen
 *$Id: audio.c,v 1.1.1.1 2008-07-23 23:30:02 vah Exp $ */

#include <stdio.h>
#include <stdlib.h>

#include "sysconf.h"
#include "data.h"
#include "myerror.h"
#include "audio.h"
#include "cache.h"

volatile bool audioPlaying = false;

Uint16 globalVolume = (AUDIO_VOLUME_NORMAL/2); // (8/8 fixed point)

/* The audio format in use (after calling initAudio().) */
static SDL_AudioSpec audiospec;

/* Sound channels */
static struct {
  const Sint16 *data; // currently playing sample data
  Uint32 length;      // data length in samples (16/16 fixed point)
  Uint32 pos;         // current playback position
  Uint32 freq;        // frequency scaling 16/16 fixed point
  Uint16 vol;         // amplitude scaling (8/8 fixed point)
  bool loop;
} channel[MIX_CHANNELS];

/* Mixer routine that SDL calls to generate sound */
static void audioCallback(void *userdata, Uint8 *stream, int len)
{
  userdata = userdata; // not used

  if(!audioPlaying) return;

  Sint16 *output = (Sint16*)stream;

  len >>= 1;

  for(unsigned s = 0; s < (unsigned)len; ++s) {
    Sint32 sample = 0;

    for(int i = 0; i < MIX_CHANNELS; ++i) {
      Uint32 length = channel[i].length;
      if(length) {
	Uint32 pos = channel[i].pos + s * channel[i].freq;

	if(pos >= length) {
	  if(channel[i].loop) {
	    while(pos >= length) 
	      pos -= length;
	  } else {
	    /* sample has ended */
	    channel[i].data = NULL;
	    channel[i].length = 0;
	    continue;
	  }
	}
    	
	sample += channel[i].vol * channel[i].data[pos >> 16];
      }
    }

    if(sample < -32768*256)
      sample = -32768;
    else if(sample > 32767*256)
      sample = 32767;
    else
      sample >>= 8;

    *output++ = (globalVolume * sample) >> 8;
  }

  bool playing = false;
  
  for(int i = 0; i < MIX_CHANNELS; ++i) {
    Uint32 length = channel[i].length;
    if(length) {
      Uint32 pos = channel[i].pos + len * channel[i].freq;
      if(pos < length) {
	/* update sample position */
	channel[i].pos = pos;
	playing = true;
      } else if(channel[i].loop) {
	/* loop sample */
	channel[i].pos = pos - length;
	playing = true;
      } else {
	/* sampled has ended */
	channel[i].data = NULL;
	channel[i].length = 0;
      }
    }
  }
  
  audioPlaying = playing;
}

static void freeAudio(void)
{
  SDL_QuitSubSystem(SDL_INIT_AUDIO);
}

void initAudio(void)
{
  CHECK_SDL_ERROR(!SDL_InitSubSystem(SDL_INIT_AUDIO));
  atexit(freeAudio);

  SDL_AudioSpec as;
  as.freq = SAMPLE_RATE;
  as.format = AUDIO_S16SYS;
  as.channels = 1;
  as.samples = AUDIO_BUFFER_SIZE;
  as.callback = audioCallback;
  as.userdata = NULL;
  CHECK_SDL_ERROR(!SDL_OpenAudio(&as, &audiospec));
  
  stopAllChannels();

  SDL_PauseAudio(0);
}

/*
 * Converts "anything" to single channel AUDIO_S16SYS. Does not change sample rate.
 * Tested with AUDIO_S16LSB and AUDIO_U8 on a little-endian computer. (Can't find examples of anything else.)
 */
static void convertSample(Uint8 *src, Uint16 format, int channels, Sint16 *dst, int length)
{
  bool eight = AUDIO_U8 == format || AUDIO_S8 == format;
  bool usign = AUDIO_U8 == format || AUDIO_U16LSB == format || AUDIO_U16MSB == format;
  bool swap = format != AUDIO_U16SYS && format != AUDIO_S16SYS;

  while(length--) {
    Sint32 sample = 0;
    
    for(int i = 0; i < channels; ++i) {
      Sint16 s;

      Uint8 a = *src++;
      if(eight) { /* 8-bit */
	if(usign) s =  a - 128;
	else s = a;
	s <<= 8;
	if(s&0x100) s |= 0xff;

      } else { /* 16-bit */
	Uint8 b = *src++;
	if(swap) s = (a << 8) | b;
	else s = (b << 8) | a;
	if(usign) s -= 0x8000;
      }
      
      sample += s;
    }

    *dst++ = sample / channels;
  }
}

static Cache *sampleCache = NULL;

static void freeSample(const char *filename, void *data)
{
  printf("\t%s\n", filename);
  Sample *sample = data;
  free(sample->data);
  free(sample);
}

static void freeSampleCache(void)
{
  SDL_LockAudio();
  stopAllChannels();
  SDL_UnlockAudio();
  puts("Free samples:");
  if(sampleCache) {
    freeCache(sampleCache, freeSample);
    sampleCache = NULL;
  }
}

const Sample* loadSample(const char *filename)
{
  if(sampleCache) {
    const Sample *sample = getFromCache(sampleCache, filename);
    if(sample) return sample;
  } else {
    sampleCache = initCache();
    atexit(freeSampleCache);
  }

  printf("Loading sample: %s\n", filename);

  SDL_AudioSpec wavspec;
  Uint8 *wavdata;
  Uint32 wavlength;
  CHECK_SDL_ERROR(SDL_LoadWAV(getFilePath(filename), &wavspec, &wavdata, &wavlength));

  switch(wavspec.format) {
  case AUDIO_U16LSB:
  case AUDIO_U16MSB:
  case AUDIO_S16LSB:
  case AUDIO_S16MSB:
    wavlength /= 2; // 2 bytes per sample
    break;
  case AUDIO_U8:
  case AUDIO_S8:
    break;
  default:
    fprintf(stderr, "Unsupported audio format\n");
    exit(1);
    break;    
  }

  Sample *sample = malloc(sizeof(Sample));
  if(!sample) {
    fprintf(stderr, "malloc failed.\n");
    exit(1);
  }
  sample->length = wavlength / wavspec.channels;
  sample->data = malloc(2*sample->length);
  if(!sample->data) {
    fprintf(stderr, "malloc failed.\n");
    exit(1);
  }
  if(wavspec.format == audiospec.format && wavspec.channels == audiospec.channels) {   
    memcpy(sample->data, wavdata, 2*sample->length);
  } else {
    convertSample(wavdata, wavspec.format, wavspec.channels, sample->data, sample->length);
  }
  SDL_FreeWAV(wavdata);
  sample->freq = (0x100 * wavspec.freq) / audiospec.freq;

  addToCache(sampleCache, filename, sample);

  return sample;
}


// SDL_LockAudio() must have been called before calling this!
static inline int findFreeChannel(void)
{  
  unsigned minleft = (unsigned)-1;
  int min = -1;
  for(int i = 0; i < MIX_CHANNELS; ++i) {    
    unsigned length = channel[i].length;
    if(!length) return i;
    if(channel[i].loop) continue; // never replace looping sounds
    length -= channel[i].pos;
    length *= 2048 - (channel[i].freq >> 8);
    if(length < minleft) { 
      minleft = length;
      min = i;
    }
  }
  return min;
}

int playSample(const Sample *sample, unsigned vol, unsigned freq, bool loop)
{
  freq = (freq * sample->freq) >> 8;

  SDL_LockAudio();

  int i = findFreeChannel();
  if(i != -1) {
    channel[i].data = sample->data;
    channel[i].length = sample->length << 16;
    channel[i].pos = 0;
    channel[i].vol = vol;
    channel[i].freq = freq;
    channel[i].loop = loop;

    audioPlaying = true;
  }
  SDL_UnlockAudio();

  return i;
}

void alterChannel(int i, const Sample *sample, unsigned vol, unsigned freq, bool loop)
{
  if(i < 0 || i >= MIX_CHANNELS) return;

  SDL_LockAudio();

  if(channel[i].data == sample->data) {
    if(vol) channel[i].vol = vol;
    if(freq) channel[i].freq = (freq * sample->freq) >> 8;    
    channel[i].loop = loop;
  }

  SDL_UnlockAudio();
}

void stopChannel(int i)
{
  if(i < 0 || i >= MIX_CHANNELS) return;
  
  SDL_LockAudio();
  
  channel[i].data = NULL;
  channel[i].length = 0;
  
  SDL_UnlockAudio();
}

void stopAllChannels(void)
{
  SDL_LockAudio();

  for(int i = 0; i < MIX_CHANNELS; ++i) {
    channel[i].data = NULL;
    channel[i].length = 0;
  }

  audioPlaying = false;

  SDL_UnlockAudio();
}

