#ifndef AUDIO_H
#define AUDIO_H
/* Copyright (C) 2008 Vin Helminen
 *$Id: audio.h,v 1.1.1.1 2008-07-23 23:30:02 vah Exp $ */

#include <stdbool.h>
#include <SDL/SDL.h>

typedef struct Sample {
  Sint16 *data;   // sample data
  Uint32 length;  // sample length
  Uint16 freq;    // frequency scaling (8/8 fixed point)
} Sample;

#define AUDIO_RATE_NORMAL     0x10000
#define AUDIO_VOLUME_NORMAL   0x100

/* mixer volume which affects all sound */
extern Uint16 globalVolume;

/* true if any sample is currently playing */
extern volatile bool audioPlaying; /* read only! */

void initAudio(void);

/* loads a sample from a .wav file
 * the samples are cached */
const Sample* loadSample(const char *filename);


/* Start playing a sample
 *
 * volume should be something between 0 and AUDIO_VOLUME_NORMAL (please, do not amplify. it will probably sound bad)
 * rate should be something around AUDIO_RATE_NORMAL (/2 means one octave lower, *2 means one octave higher, etc.)
 * if loop is true the sound loops forever, otherwise, plays just once
 *
 * Returns the channel number (>=0) the sample is played on or -1 if there wasn't any free channels left.
 * 
 * Note that looped sounds reserve the sound channel for themselves until stopped with stopChannel()
 * and unlooped sounds can be stopped at any time to free the channel for another sound.
 */
int playSample(const Sample *sample, unsigned volume, unsigned rate, bool loop);

/* A macro that simplifies the above if you just need to play an simple sound effect */
#define play(sample) playSample(sample, AUDIO_VOLUME_NORMAL, AUDIO_RATE_NORMAL, false)

/* Alter a currently playing sample.
 *
 * channel is the channel number returned by playSample()
 * sample is the same sample that was started with playSample() on that channel (this is enforced)
 * volume is the new volume or 0 to leave unchanged (sorry, cannot therefore mute the channel completely)
 * rate is the new rate or 0 to leave unchanged (for pitch bend style effects)
 * loop can be used to stop looping if set to false
 *
 * It might be a good idea to use this only for looping sounds.
 */ 
void alterChannel(int channel, const Sample *sample, unsigned volume, unsigned rate, bool loop);

/* Stop whatever sample is playing on the given channel.
 *
 * It is a good idea to only use this with looping sounds.
 */
void stopChannel(int channel);

/* Stop all sound on all channels. Later just call playSample() start playing new samples. */
void stopAllChannels(void);


#endif
