/*
 * Copyright (C) 2006-2008 Vin Helminen 
 * $Id: game.c,v 1.1.1.1 2008-07-23 23:30:02 vah Exp $
 */

#include <stdbool.h>

#include "game.h"
#include "myerror.h"
#include "video.h"
#include "audio.h"
#include "data.h"


#define MAX_SEQUENCE_LENGTH 99

enum { A = 0, B, X, Y, GAMEOVER };

static SDL_Surface *background = NULL;
static const Sprite *lights = NULL, *numbers = NULL;
static const Sample *sounds[5] = {NULL};

static enum {INTRO, COMPUTER, PLAYER, PL2CPU} state;
static short sequence[MAX_SEQUENCE_LENGTH];
static int seqPos, seqLen;
static bool lightState[4];

static int counter;

static inline void freeImages(void)
{
  puts("Free images:");
  if(background) {
    puts("\t" IMG_BACKGROUND);
    SDL_FreeSurface(background);
    background = NULL;
  } 
}


static inline void loadData(void)
{
  puts("Loading image: " IMG_BACKGROUND);
  background = loadImage(IMG_BACKGROUND, 0, 0);
  atexit(freeImages);

  lights = loadSprites(SPR_LIGHTS, 0);
  numbers = loadSprites(SPR_NUMBERS, 0);

  sounds[A] = loadSample(SND_A);
  sounds[B] = loadSample(SND_B);
  sounds[X] = loadSample(SND_X);
  sounds[Y] = loadSample(SND_Y);
  sounds[GAMEOVER] = loadSample(SND_GAMEOVER);
}

void initGame(void)
{
  loadData();
  lightState[A] = lightState[B] = lightState[X] = lightState[Y] = false;
  state = INTRO;
  counter = seqPos = seqLen = 0;
}

void gameLogic(const GameInput *input, unsigned step)
{
  counter -= step;
  if(counter < 0) counter = -1;

  switch(state) {
  default:
  case INTRO:
    {
      if(input->a || input->b || input->x || input->y) {
	/* start game */
	lightState[A] = lightState[B] = lightState[X] = lightState[Y] = true;
	state = COMPUTER;
	seqPos = 0;
	seqLen = 1;
	counter = 2;
      } else {
	/* blink lights randomly */
	int i = rand() & 3;
	lightState[i] = !lightState[i];
      }
    }
    break;

  case COMPUTER:
    {
      if(counter < 0) {
	counter = 2;
	if(lightState[A] || lightState[B] || lightState[X] || lightState[Y]) {
	  lightState[A] = lightState[B] = lightState[X] = lightState[Y] = false;
	} else {
	  if(seqPos < seqLen-1) {
	    /* play back */
	    int i = sequence[seqPos++];
	    lightState[i] = true;
	    play(sounds[i]);
	  } else if(seqPos == seqLen-1) {
	    /* add new */
	    int i = (rand()>>9) & 3;
	    sequence[seqPos++] = i;
	    lightState[i] = true;
	    play(sounds[i]);
	  } else {
	    /* give turn to the player */
	    state = PLAYER;
	    seqPos = 0;
	  }
	}
      }
    }
    break;
  case PLAYER:
    {
      int pressed = (input->a & 1) | ((input->b & 1)<<1) | ((input->x & 1)<<2) | ((input->y)<<3);

      if(pressed || counter < 0)
	lightState[A] = lightState[B] = lightState[X] = lightState[Y] = false;

      if(pressed) {
	int i = sequence[seqPos];
	if(pressed == (1 << i)) {
	  /* correct */
	  lightState[i] = true;
	  play(sounds[i]);

	  seqPos++;
	  if(seqPos == seqLen) {
	    state = PL2CPU;
	  }
	} else {
	  /* wrong */
	  lightState[A] = lightState[B] = lightState[X] = lightState[Y] = true;
	  play(sounds[GAMEOVER]);
	  state = INTRO;
	}
	counter = 2;
      }
    }
    break;

  case PL2CPU:
    if(counter < 0) {
      if(seqPos == MAX_SEQUENCE_LENGTH) {
	/* player has won */
	state = INTRO;
      } else {
	state = COMPUTER;
	seqPos = 0;
	seqLen++;
      }
    }
      
  }
}

void drawGame(SDL_Surface *surface)
{
  SDL_BlitSurface(background, NULL, surface, NULL);

  /* lights */
  for(int i = 0; i < 4; ++i) {
    if(lightState[i])
      drawSprite(lights, i, 0, 0, surface);
  }   

  /* numbers : "pos / len" */
  {
    int i;
    i = seqPos / 10;
    if(i) drawSprite(numbers, i, SPR_NUMBERS_X, SPR_NUMBERS_Y, surface);
    i = seqPos % 10;
    drawSprite(numbers, i, SPR_NUMBERS_X+SPR_NUMBERS_SPACING, SPR_NUMBERS_Y, surface);
    
    drawSprite(numbers, 10, SPR_NUMBERS_X+2*SPR_NUMBERS_SPACING, SPR_NUMBERS_Y, surface);
    
    i = seqLen / 10;
    if(i) drawSprite(numbers, i, SPR_NUMBERS_X+3*SPR_NUMBERS_SPACING, SPR_NUMBERS_Y, surface);
    i = seqLen % 10;
    drawSprite(numbers, i, SPR_NUMBERS_X+4*SPR_NUMBERS_SPACING, SPR_NUMBERS_Y, surface);
  }

}
