/*
 * Simon Game
 * Copyright (C) 2006-2008 Vin Helminen
 * $Id: main.c,v 1.4 2008-07-25 16:37:59 vah Exp $
 */

#include <stdio.h>
#include <stdlib.h>
#include <sys/time.h>

#include "myerror.h"
#include "video.h"
#include "audio.h"
#include "time.h"
#include "game.h"
#include "data.h"


#ifdef GP2X

#include "gp2x.h"

static SDL_Joystick *joystick = NULL;

static void closeJoystick(void)
{
  SDL_JoystickClose(joystick);
}
#endif

static inline void randomize(void)
{
  struct timeval tv;
  gettimeofday(&tv, NULL);
  srand((tv.tv_usec) | (tv.tv_sec));
}


static bool handleEvents(GameInput *input)
{
  static int volAdj = 0;

  SDL_Event event;

  // reset trigger values
  input->a = input->b = input->x = input->y =
    input->fullscreen = input->volume = false;

  while(SDL_PollEvent(&event)) {
    switch(event.type) {
    case SDL_QUIT:
      return false;

#ifdef GP2X

    case SDL_JOYBUTTONDOWN:
      switch(event.jbutton.button) {
      case GP2X_BUTTON_START:
	return false;

      case GP2X_BUTTON_A:
	input->a = true;
	break;
      case GP2X_BUTTON_B:
	input->b = true;
	break;
      case GP2X_BUTTON_X:
	input->x = true;
	break;
      case GP2X_BUTTON_Y:
	input->y = true;
	break;

      case GP2X_BUTTON_VOLUP:
	volAdj = 5;
	break;
      case GP2X_BUTTON_VOLDOWN:
	volAdj = -5;
	break;

      default:
	break;
      }
      break;

    case SDL_JOYBUTTONUP:
      switch(event.jbutton.button) {
      case GP2X_BUTTON_VOLUP:
      case GP2X_BUTTON_VOLDOWN:
	volAdj = 0;
	break;

      default:
	break;
      }
      break;

#else
    case SDL_KEYDOWN:
      switch(event.key.keysym.sym) {
      case SDLK_ESCAPE:
	return false;
      	
      case SDLK_UP:
	input->y = true;
	break;
      case SDLK_DOWN:
	input->x = true;
	break;
      case SDLK_LEFT:
	input->a = true;
	break;
      case SDLK_RIGHT:
	input->b = true;
	break;

      case SDLK_p:
	input->pause = !input->pause;
	break;
      case SDLK_f:
	input->fullscreen = true;
	break;

      case SDLK_PLUS:
      case SDLK_KP_PLUS:
	volAdj = 5;
	break;
      case SDLK_MINUS:
      case SDLK_KP_MINUS:
	volAdj = -5;
	break;

      default:
	break;
      }
      break;

    case SDL_KEYUP:
      switch(event.key.keysym.sym) {
      case SDLK_PLUS:
      case SDLK_KP_PLUS:
      case SDLK_MINUS:
      case SDLK_KP_MINUS:
	volAdj = 0;
	break;

      default:
	break;
      }
      break;

#endif
    }
  }

  if(volAdj) {
    int volume = globalVolume + volAdj;
    if(volume < 0)
      volume = 0;
    else if(volume > AUDIO_VOLUME_NORMAL) 
      volume = AUDIO_VOLUME_NORMAL;
    globalVolume = volume;
    input->volume = true;
  }

  return true;
}

static inline void splashScreen(SDL_Surface *screen)
{
  puts("Loading image: " IMG_SPLASH);
  SDL_Surface *splash = loadImage(IMG_SPLASH, 0, 0);

  SDL_BlitSurface(splash, NULL, screen, NULL);
  SDL_Flip(screen);

  puts("Free image: " IMG_SPLASH);
  SDL_FreeSurface(splash);
}

static inline void waitSplash(int frames)
{
  while(frames > 0) {
    SDL_Event event;
    while(SDL_PollEvent(&event)) {
      if(event.type == SDL_QUIT) exit(0);
    }
    frames -= tick();
  }
}


static inline void drawVolumeBar(SDL_Surface *surface)
{
  SDL_Rect r;
  r.x = (SCREEN_WIDTH - 1 - AUDIO_VOLUME_NORMAL)/2;
  r.y = SCREEN_HEIGHT - 21;
  r.h = 12;
  r.w = AUDIO_VOLUME_NORMAL + 2;
  SDL_FillRect(surface, &r, 0);
  r.x += 1;
  r.y += 1;
  r.h = 10;
  unsigned vol = globalVolume;
  r.w = vol;
  int c = (255*vol)/AUDIO_VOLUME_NORMAL;
  SDL_FillRect(surface, &r, SDL_MapRGB(surface->format, c, 256-c, 0));
}

int main(void)
{
  puts("Simon (C) 2006-2008 Vin Helminen");

#ifdef GP2X

  GP2XInit();
  CHECK_SDL_ERROR(!SDL_Init(SDL_INIT_JOYSTICK));
  joystick = SDL_JoystickOpen(0);
  CHECK_SDL_ERROR(joystick);
  atexit(closeJoystick);

  SDL_Surface *screen = initVideo(START_FULLSCREEN);

#else

  CHECK_SDL_ERROR(!SDL_Init(SDL_INIT_EVENTTHREAD));
  atexit(SDL_Quit);

  bool fullscreen = START_FULLSCREEN;
  SDL_Surface *screen = initVideo(fullscreen);

  SDL_WM_SetCaption("Simon", 0);

#endif
  SDL_ShowCursor(SDL_DISABLE);  

  splashScreen(screen);
  initTime(10);
  initAudio();
  initGame();
  randomize();
  waitSplash(10);

  GameInput input = {false, false, false, false, false, false, false};
  int viewVolume = 0;

  while(handleEvents(&input)) {
#ifndef GP2X
    if(input.fullscreen)
      screen = initVideo((fullscreen=!fullscreen));
#endif

    unsigned step = tick();

    gameLogic(&input, step);

    drawGame(screen);

    if(input.volume)
      viewVolume = 5;
    if(viewVolume > 0) {
      viewVolume -= step;
      drawVolumeBar(screen);
    }

    flip();
  }

  return EXIT_SUCCESS;
}
