/*
 * Copyright (C) 2006-2008 Vin Helminen 
 * $Id: video.c,v 1.2 2008-07-23 23:50:49 vah Exp $
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdbool.h>

#include <SDL/SDL_image.h>

#include "myerror.h"
#include "data.h"
#include "video.h"
#include "cache.h"

void SetPalette(SDL_Surface *surface)
{
  SDL_Color c[256];
  for(unsigned i = 0; i < 256; ++i) {
    unsigned t = i>>5;
    c[i].r = (t<<5) | (t<<2) | (t>>1);
    t = (i>>2)&7;
    c[i].g = (t<<5) | (t<<2) | (t>>1);
    t = i & 3;
    t = t | (t<<4);
    c[i].b = t | (t<<2);
  }
  SDL_SetColors(surface, c, 0, 256);
}

/* SCREEN MANAGEMENT */

static SDL_Surface *screen = NULL;

static void freeVideo(void)
{
  if(screen) {
#ifdef GP2X
    /* clear screen (twice for double buffering) */
    SDL_FillRect(screen, NULL, 0);
    SDL_Flip(screen);
    SDL_FillRect(screen, NULL, 0);
    SDL_Flip(screen);
#else
    SDL_QuitSubSystem(SDL_INIT_VIDEO);
#endif
    screen = NULL;
  }
}

SDL_Surface* initVideo(bool fullscreen)
{
  if(!SDL_WasInit(SDL_INIT_VIDEO)) {
    CHECK_SDL_ERROR(!SDL_InitSubSystem(SDL_INIT_VIDEO));
    atexit(freeVideo);
  }

  Uint32 flags = SDL_HWSURFACE|SDL_DOUBLEBUF|SDL_ANYFORMAT;

  if(fullscreen)
    flags |= SDL_FULLSCREEN;

  screen = SDL_SetVideoMode(SCREEN_WIDTH, SCREEN_HEIGHT, SCREEN_BPP, flags);
  CHECK_SDL_ERROR(screen);
  if(screen->format->BytesPerPixel == 1)
    SetPalette(screen);

  return screen;
}

void flip(void)
{
  SDL_Flip(screen);
}

SDL_Surface* loadImage(const char *filename, int width, int height)
{
  SDL_Surface *loaded = IMG_Load(getFilePath(filename));
  CHECK_SDL_ERROR(loaded);

  if(width <= 0) width = loaded->w;
  if(height <= 0) height = loaded->h;

  SDL_Surface *converted;
  if(width != loaded->w || height != loaded->h) {
    /* need to scale the image */
    converted = SDL_CreateRGBSurface(SDL_HWSURFACE, width, height,
				     screen->format->BitsPerPixel,
				     screen->format->Rmask,
				     screen->format->Gmask,
				     screen->format->Bmask,
				     screen->format->Amask);
    CHECK_SDL_ERROR(converted);
    scaleSurface(loaded, converted);
  } else {
    /* no need for scaling */
    converted = SDL_ConvertSurface(loaded, screen->format, SDL_HWSURFACE);
    CHECK_SDL_ERROR(converted);
  }

  SDL_FreeSurface(loaded);

  return converted;
}

/* Sprites and text */

static Cache *spriteCache = NULL;

static void freeSprite(const char *filename, void *data)
{
  printf("\t%s\n", filename);
  Sprite *sprite = data;
  SDL_FreeSurface(sprite->surface);
  free(sprite);
}

static void freeSprites(void)
{
  puts("Free sprites:");
  if(spriteCache) {
    freeCache(spriteCache, freeSprite);
    spriteCache = NULL;
  }
}

const Sprite* loadSprites(const char *filename, int width, int height, int offset)
{
  if(spriteCache) {
  /* try to find an already loaded texture */
    Sprite *sprite = getFromCache(spriteCache, filename);
    if(sprite) return sprite;
  } else {
    spriteCache = initCache();
    atexit(freeSprites);
  }

  /* load new sprite */
  printf("Loading sprites: %s\n", filename);

  SDL_Surface *img = loadImage(filename, 0, 0);
  // no need to check for success here

  CHECK_SDL_ERROR(!SDL_SetColorKey(img, SDL_SRCCOLORKEY|SDL_RLEACCEL, SDL_MapRGB(img->format, 0xff, 0, 0xff)));

  Sprite *sprite;
  sprite = malloc(sizeof *sprite);
  CHECK_MALLOC(sprite);
  sprite->surface = img;
  sprite->width = width;
  sprite->height = height;
  sprite->offset = offset;

  addToCache(spriteCache, filename, sprite);
  return sprite;
}

void drawSprite(const Sprite *sprite, int index, int x, int y, SDL_Surface *surface)
{
  div_t col = div((index - sprite->offset) * sprite->width, sprite->surface->w);
  SDL_Rect srcrect, dstrect;
  srcrect.x = col.rem;
  srcrect.y = sprite->height * col.quot;
  srcrect.w = sprite->width;
  srcrect.h = sprite->height;
  dstrect.x = x;
  dstrect.y = y;
  SDL_BlitSurface(sprite->surface, &srcrect, surface, &dstrect);
}

void drawString(const Sprite *font, const char *text, int x, int y, SDL_Surface *surface)
{
  if(x < 0) x += surface->w - font->width*strlen(text);
  int c;
  while((c = *text++)) {
    drawSprite(font, c, x, y, surface);
    x += font->width;
  }
}

/* Preprocessing */

/* 32 bits per pixel */
static void setPixel32(SDL_Surface *s, int x, int y, SDL_Color c)
{
  ((Uint32*)s->pixels)[x + y*(s->pitch/4)] = SDL_MapRGB(s->format, c.r, c.g, c.b);
}

static SDL_Color getPixel32(const SDL_Surface *s, int x, int y)
{
  SDL_Color c;
  SDL_GetRGB(((Uint32*)s->pixels)[x + y*(s->pitch/4)], s->format, &c.r, &c.g, &c.b);
  return c;
}

/* 24 bits per pixel */
static void setPixel24(SDL_Surface *s, int x, int y, SDL_Color c)
{
  Uint32 v = SDL_MapRGB(s->format, c.r, c.g, c.b);
  Uint8* p = ((Uint8 *)s->pixels) + 3*x + y*s->pitch;
#if SDL_BYTEORDER == SDL_BIG_ENDIAN
  p[0] = (v >> 16) & 0xff;
  p[1] = (v >> 8) & 0xff;
  p[2] = v & 0xff;
#else
  p[0] = v & 0xff;
  p[1] = (v >> 8) & 0xff;
  p[2] = (v >> 16) & 0xff;
#endif
}

static SDL_Color getPixel24(const SDL_Surface *s, int x, int y)
{
  SDL_Color c;
  Uint8* p = ((Uint8*)s->pixels) + 3*x + y*s->pitch;
  Uint32 v;
#if SDL_BYTEORDER == SDL_BIG_ENDIAN
  v = p[0] << 16 | p[1] << 8 | p[2];
#else
  v = p[0] | p[1] << 8 | p[2] << 16;
#endif
  SDL_GetRGB(v, s->format, &c.r, &c.g, &c.b);
  return c;
}

/* 16 bits per pixel */
static void setPixel16(SDL_Surface *s, int x, int y, SDL_Color c)
{
  ((Uint16*)s->pixels)[x + y*(s->pitch/2)] = SDL_MapRGB(s->format, c.r, c.g, c.b);
}

static SDL_Color getPixel16(const SDL_Surface *s, int x, int y)
{
  SDL_Color c;
  SDL_GetRGB(((Uint16*)s->pixels)[x + y*(s->pitch/2)], s->format, &c.r, &c.g, &c.b);
  return c;
}

/* 8 bits per pixel */
static void setPixel8(SDL_Surface *s, int x, int y, SDL_Color c)
{
  ((Uint8*)s->pixels)[x + y*s->pitch] = SDL_MapRGB(s->format, c.r, c.g, c.b);
}

static SDL_Color getPixel8(const SDL_Surface *s, int x, int y)
{
  SDL_Color c;
  SDL_GetRGB(((Uint8*)s->pixels)[x + y*s->pitch], s->format, &c.r, &c.g, &c.b);
  return c;
}


SetPixelFunc getSetPixel(SDL_Surface *surface)
{
  switch(surface->format->BytesPerPixel) {
  default:
    ERROR(255, "Invalid BytesPerPixel");
  case 1:
    return setPixel8;
  case 2:
    return setPixel16;
  case 3:
    return setPixel24;
  case 4:
    return setPixel32;
  };
}

GetPixelFunc getGetPixel(const SDL_Surface *surface)
{
  switch(surface->format->BytesPerPixel) {
  default:
    ERROR(255, "Invalid BytesPerPixel");
  case 1:
    return getPixel8;
  case 2:
    return getPixel16;
  case 3:
    return getPixel24;
  case 4:
    return getPixel32;
  };
}

void scaleSurface(const SDL_Surface *src, SDL_Surface *dst)
{
  if(src->w == dst->w && src->h == dst->h) {
    /* dimensions match. no need for scaling */
    SDL_BlitSurface((SDL_Surface*)src, NULL, dst, NULL);
    return;
  }

  if(SDL_MUSTLOCK(src)) SDL_LockSurface((SDL_Surface*)src);
  if(SDL_MUSTLOCK(dst)) SDL_LockSurface(dst);

  GetPixelFunc get = getGetPixel(src);
  SetPixelFunc set = getSetPixel(dst);

  bool aaw = src->w > dst->w, aah = src->h > dst->h;
  Uint32 scalex = ((src->w)<<16)/dst->w, scaley = (src->h<<16)/dst->h;
  for(Uint32 dy = 0, sy = 0; dy < (unsigned)dst->h; ++dy, sy += scaley) {
    Uint32 tmpy = sy >> 16;
    for(Uint32 dx = 0, sx = 0; dx < (unsigned)dst->w; ++dx, sx += scalex) {
      Uint32 tmpx = sx >> 16;
      SDL_Color c1 = get(src, tmpx, tmpy);
      if(aaw && aah) {
	SDL_Color c2, c3, c4;
	Uint32 tmpx2 = (sx + scalex/2) >> 16, tmpy2 = (sy + scaley/2) >> 16;
	if(tmpy2 < (unsigned)src->h) {
	  c2 = get(src, tmpx, tmpy2);
	  if(tmpx2 < (unsigned)src->w) {
	    c3 = get(src, tmpx2, tmpy);
	    c4 = get(src, tmpx2, tmpy2);
	  } else {
	    c3 = c1;
	    c4 = c2;
	  }
	} else {
	  c2 = c1;
	  if(tmpx2 < (unsigned)src->w)
	    c3 = c4 = get(src, tmpx2, tmpy);
	  else
	    c3 = c4 = c1;
	}
	c1.r = (c1.r + c2.r + c3.r + c4.r)>>2;
	c1.g = (c1.g + c2.g + c3.g + c4.g)>>2;
	c1.b = (c1.b + c2.b + c3.b + c4.b)>>2;
      } else if(aaw) {
	Uint32 tmpx2 = (sx + scalex/2) >> 16;
	if(tmpx2< (unsigned)src->w) {
	  SDL_Color c2 = get(src, tmpx2, tmpy);
	  c1.r = (c1.r + c2.r)>>1;
	  c1.g = (c1.g + c2.g)>>1;
	  c1.b = (c1.b + c2.b)>>1;
	}	  
      } else if(aah) {
	Uint32 tmpy2 = (sy + scaley/2) >> 16;
	if(tmpy2 < (unsigned)src->h) {
	  SDL_Color c2 = get(src, tmpx, tmpy2);
	  c1.r = (c1.r + c2.r)>>1;
	  c1.g = (c1.g + c2.g)>>1;
	  c1.b = (c1.b + c2.b)>>1;
	}
      }
      set(dst, dx, dy, c1);
    }
  }

  if(SDL_MUSTLOCK(src)) SDL_UnlockSurface((SDL_Surface*)src);
  if(SDL_MUSTLOCK(dst)) SDL_UnlockSurface(dst);
}
